package mvd.jester.simulator;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.AdditionalMatchers.gt;
import static org.mockito.AdditionalMatchers.lt;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import mvd.jester.model.Task;
import mvd.jester.simulator.internals.ProcessorContext;
import mvd.jester.simulator.internals.TaskContextInterface;
import mvd.jester.simulator.internals.maiabertogna.JobContext;
import mvd.jester.simulator.internals.maiabertogna.TaskContext;

/**
 * TestProcessorContext
 */
public class TestProcessorContext {

    @Test
    @DisplayName("Check if the job execution is updated correctly.")
    void testUpdateExecution() {
        for (int run = 0; run < 100; ++run) {
            long jobWcet = ThreadLocalRandom.current().nextLong(1, 10);

            TaskContext task = mock(TaskContext.class);
            JobContext job = mock(JobContext.class);
            when(job.updateExecution(lt(jobWcet))).thenReturn(Optional.empty());
            when(job.updateExecution(jobWcet)).thenReturn(Optional.of(task));
            when(job.updateExecution(gt(jobWcet))).thenThrow(RuntimeException.class);

            ProcessorContext processor = new ProcessorContext(1);
            assertFalse(processor.updateExecution(0).isPresent());
            processor.setJob(job);

            for (int i = 0; i < jobWcet; ++i) {
                assertFalse(processor.updateExecution(i).isPresent());
            }
            Optional<TaskContextInterface> tci = processor.updateExecution(jobWcet);
            assertTrue(tci.isPresent());
            assertTrue(tci.get().equals(task));

            assertThrows(RuntimeException.class, () -> processor.updateExecution(jobWcet + 1));

            verify(job, times((int) jobWcet + 2)).updateExecution(anyLong());
        }
    }


    @Test
    @DisplayName("Check if the processor correctly accepts jobs")
    void testAcceptTask() {
        for (int run = 0; run < 100; ++run) {
            long firstPeriod = ThreadLocalRandom.current().nextLong(100, 1000);
            long secondPeriod = ThreadLocalRandom.current().nextLong(100, 1000);

            JobContext firstJob = mock(JobContext.class);
            TaskContext firstTaskContext = mock(TaskContext.class);
            Task firstTask = mock(Task.class);

            when(firstJob.getTaskContext()).thenReturn(firstTaskContext);
            when(firstJob.prepareJob(anyLong())).thenReturn(true);
            when(firstTaskContext.getTask()).thenReturn(firstTask);
            when(firstTaskContext.getNextJob()).thenReturn(Optional.of(firstJob));
            when(firstTask.getPeriod()).thenReturn((long) firstPeriod);


            JobContext secondJob = mock(JobContext.class);
            TaskContext secondTaskContext = mock(TaskContext.class);
            Task secondTask = mock(Task.class);

            when(secondJob.getTaskContext()).thenReturn(secondTaskContext);
            when(secondJob.prepareJob(anyLong())).thenReturn(true);
            when(secondTaskContext.getTask()).thenReturn(secondTask);
            when(secondTaskContext.getNextJob()).thenReturn(Optional.of(secondJob));
            when(secondTask.getPeriod()).thenReturn((long) secondPeriod);

            ProcessorContext processor = new ProcessorContext(1);

            assertTrue(processor.acceptTask(secondTaskContext, run));
            if (firstPeriod < secondPeriod) {
                assertTrue(processor.acceptTask(firstTaskContext, run));
            } else {
                assertFalse(processor.acceptTask(firstTaskContext, run));
            }
            assertFalse(processor.acceptTask(secondTaskContext, run));

            int time = firstPeriod < secondPeriod ? 1 : 0;
            verify(firstJob, times(time)).prepareJob(anyLong());
            verify(firstJob, times(time)).setCurrentProcessor(processor);
            verify(firstTaskContext, times(time)).getNextJob();
            verify(secondJob, times(1)).prepareJob(anyLong());
            verify(secondJob, times(1)).setCurrentProcessor(processor);
            verify(secondJob, times(time)).setCurrentProcessor(null);
            verify(secondTaskContext, times(1)).getNextJob();
        }
    }

}
